from aflow.control import search

_keywords = []
"""list of `str` keyword names available in AFLOW.
"""
def list_keywords():
    """Returns a list of all possible keywords available in the AFLOW
    lib.
    """
    global _keywords
    if len(_keywords) == 0:
        from aflow.keywords import Keyword
        import aflow.keywords as kw
        from inspect import getmembers
        
        for name, o in getmembers(kw):
            if isinstance(o, Keyword):
                _keywords.append(name)

    return _keywords

import aflow.keywords as K



"""Documentation and format builders for the latest AFLUX
specification. Since AFLUX is in constant flux, we generate the
documentation and keyword lists dynamically at development time
whenever a new version is released. This happens by parsing the HTML
page that lists all available keywords and then querying the `help`
feature of the AFLOW API for additional information. This is then used
to generate the API documentation automatically.
"""
from six.moves import urllib
def _get_keywords():
    """Returns a list of the currently valid keywords for the AFLUX API.
    """
    from six.moves import urllib
    import json
    urlopen = urllib.request.urlopen
    url = "http://aflowlib.duke.edu/search/API/?schema"
    r = urlopen(url).read().decode("utf-8")
    return json.loads(r)

def _get_kw_help(keyword, metadata):
    """Gets the documentation string and type information, etc. from
    the AFLUX API.

    Args:
        keyword (str): name of the keyword to get help for.
        metadata (dict): key-value pairs describing the keyword's
          metadata in the AFLOW database.

    Returns:
        dict: with key-value pairs returned by AFLUX (if the entry
        exists); otherwise `None`.
    """
    _set_defaults(metadata)

    from aflow.caster import ptype, docstrings
    metadata["ptype"] = ptype(metadata["type"], keyword)
    if keyword in docstrings:
        metadata["customdoc"] = docstrings[keyword]
    
    return metadata

def _set_defaults(d):
    """Adds the default values to the specified dictionary if they don't already
    exist. This includes all the parameters necessary to construct the templates
    for the classes.

    .. note:: This mutates the given dictionary if some of the keys are missing.
    """
    d.setdefault("title", "No schema information available.")
    d.setdefault("inclusion", "unknown")
    d.setdefault("status", "unknown")
    d.setdefault("description", "No description was returned from AFLUX.")
    d.setdefault("type")

def keywords(root=None):
    """Generates a python module file for the :class:`~aflow.entries.Entry` that
    has documented methods for lazy loading of properties from AFLOW database.

    Args:
        root (str): path in which to generate the module files.
    """
    from os import path
    from aflow.utility import reporoot
    if root is None: # pragma: no cover
        entries = path.join(reporoot, "aflow", "entries.py")
        keyword = path.join(reporoot, "aflow", "keywords.py")
    else:
        entries = path.join(root, "entries.py")
        keyword = path.join(root, "keywords.py")

    #Compile a dictionary of all the keywords and their corresponding
    #dictionaries.
    from collections import OrderedDict
    kws = _get_keywords()
    kwdata = OrderedDict()
    for kw in sorted(kws.keys()):
        metadata = kws[kw]
        if kw[0:2] != "__":
            kwdata[kw] = _get_kw_help(kw, metadata)
        
    settings = {
        "keywords": kwdata
    }

    from jinja2 import Environment, PackageLoader
    env = Environment(loader=PackageLoader('aflow', 'templates'),
                      keep_trailing_newline=True)
    
    tentry = env.get_template("entry.py")
    with open(entries, 'w') as f:
        f.write(tentry.render(**settings))

    tkeyword = env.get_template("keywords.py")
    with open(keyword, 'w') as f:
        f.write(tkeyword.render(**settings))



"""Implements classes to represent each keyword with overloaded
operators to make querying with AFLUX intuitive.
"""
from six import string_types
import numpy

_all_keywords = []
"""list: of `str` keyword names for which class instances exist within
this module.
"""

def load(target):
    """Loads all keywords into the specified target dictionary.
    """
    import sys
    self = sys.modules[__name__]    
    _find_all()

    for n in _all_keywords:
        cls = getattr(self, n)
        target[n] = cls

def _find_all():
    """Finds the names of all keywords supported in this module.
    """
    #Get a reference to the module and its global keyword cache.
    global _all_keywords    
    if len(_all_keywords) == 0:
        import sys
        import inspect
        self = sys.modules[__name__]
        for n, o in inspect.getmembers(self):
            if isinstance(o, Keyword):
                _all_keywords.append(n)    

def reset():
    """Resets all the keyword instances internal states so that they
    can be re-used for a new query.
    """
    import sys
    self = sys.modules[__name__]
    _find_all()
    
    for n in _all_keywords:
        cls = getattr(self, n)
        cls.state = []
        cls.cache = []

class Keyword(object):
    """Represents an abstract keyword that can be sub-classed for a
    specific material attribute. This class also represents logical
    operators that define search queries. The combination of two
    keywords with a logical operator produces one more keyword, but
    which has its :attr:`state` altered.

    Args:
        state (str): current query state of this keyword (combination).

    Attributes:
        state (list): of `str` *composite* queries for this keyword (combination).
        ptype (type): python type that values for this keyword will have.
        name (str): keyword name to use in the AFLUX request.
        cache (list): of `str` *simple* operator comparisons.
        classes (set): of `str` keyword names that have been combined into the
          current keyword.
    """
    name = None
    ptype = None
    atype = None
    
    def __init__(self, state=None):
        self.state = state if state is not None else []
        self.cache = []
        self.classes = set([self.name])

    def __hash__(self):
        return hash(self.name)
        
    def __str__(self):
        if len(self.state) == 1:
            s = self.state[0]
        elif len(self.cache) == 1:
            s = self.cache[0]
        else:
            return self.name

        if len(self.classes) == 1:
            return "{0}({1})".format(self.name, s)
        else:
            return s
        
    def __lt__(self, other):
        if isinstance(other, string_types):
            self.cache.append("*'{0}'".format(other))
        else:
            self.cache.append("*{0}".format(other))
        return self
            
    def __gt__(self, other):
        if isinstance(other, string_types):
            self.cache.append("'{0}'*".format(other))
        else:
            self.cache.append("{0}*".format(other))
        return self

    def __mod__(self, other):
        assert isinstance(other, string_types)
        self.cache.append("*'{0}'*".format(other))
        return self
    
    def __eq__(self, other):
        if isinstance(other, string_types):
            self.cache.append("'{0}'".format(other))
        else:
            self.cache.append("{0}".format(other))
        return self

    def _generic_combine(self, other, token):
        if other is self:
            #We need to do some special handling. We shouldn't have
            #more than two entries in cache; otherwise something went
            #wrong.
            if len(self.cache) == 2:
                args = self.cache[0], token, self.cache[1]
                self.state.append("{0}{1}{2}".format(*args))
            elif len(self.cache) == 1 and len(self.state) == 1:
                args = self.cache[0], token, self.state[0]
                self.state = ["{0}{1}({2})".format(*args)]
            elif len(self.state) == 2:
                args = self.state[0], token, self.state[1]
                self.state = ["({0}){1}({2})".format(*args)]
            else:
                raise ValueError("Inconsistent operators; check your parenthesis.")
                
            self.cache = []
            return self
        else:
            #Just combine the two together into a new keyword that has
            #the combined state.
            s = None
            if len(self.state) == 1 and len(self.cache) == 0:
                s = self.state[0]
            elif len(self.state) == 0 and len(self.cache) == 1:
                s = self.cache[0]
            #if ':' in s or ',' in s:
            s = "{0}({1})".format(self.name, s)
                
            o = None
            if len(other.state) == 1 and len(other.cache) == 0:
                o = other.state[0]
            elif len(other.state) == 0 and len(other.cache) == 1:
                o = other.cache[0]
            #if ':' in o or ',' in o:
            o = "{0}({1})".format(other.name, o)

            assert s is not None
            assert o is not None            
            result = Keyword(["{0}{1}{2}".format(s, token, o)])
            result.classes = self.classes | other.classes
            return result
    
    def __and__(self, other):
        return self._generic_combine(other, ',')
        
    def __or__(self, other):
        return self._generic_combine(other, ':')

    def __invert__(self):
        assert len(self.state) == 1 or len(self.cache) == 1
        #The two uncovered cases below are coded for completeness, but we don't
        #have any tests to trigger them; i.e., they don't come up in normal
        #operations.
        if len(self.state) == 1:
            if '!' in self.state[0]:
                self.state[0] = self.state[0].replace('!', '')
            elif '(' in self.state[0]:
                self.state[0] = self.state[0].replace('(', "(!")
            else:# pragma: no cover
                self.state[0] = '!' + self.state[0]
        elif len(self.cache) == 1:
            if '!' in self.cache[0]:
                self.cache[0] = self.cache[0].replace('!', '')
            elif '(' in self.cache[0]:# pragma: no cover
                self.cache[0] = self.cache[0].replace('(', "(!")
            else:
                self.cache[0] = '!' + self.cache[0]
        return self    
    
    
class _Bravais_lattice_orig(Keyword):
    """original bravais lattice (`optional`). Units: ``.
    
    

    Returns:
        str: Returns the Bravais lattice of the original unrelaxed structure before the calculation.
        
    """
    name = "Bravais_lattice_orig"
    ptype = str
    atype = "string"

Bravais_lattice_orig = _Bravais_lattice_orig()
    
class _Bravais_lattice_relax(Keyword):
    """relaxed bravais lattice (`optional`). Units: ``.
    
    

    Returns:
        str: Returns the Bravais lattice of the original relaxed structure after the calculation.
        
    """
    name = "Bravais_lattice_relax"
    ptype = str
    atype = "string"

Bravais_lattice_relax = _Bravais_lattice_relax()
    
class _Egap(Keyword):
    """electronic energy band gap (`mandatory`). Units: `eV`.
    
    

    Returns:
        float: Band gap calculated with the approximations and pseudopotentials described by other keywords.
        
    """
    name = "Egap"
    ptype = float
    atype = "number"

Egap = _Egap()
    
class _Egap_fit(Keyword):
    """fitted band gap (`mandatory`). Units: `eV`.
    
    

    Returns:
        float: Simple cross-validated correction (fit) of Egap.
        
    """
    name = "Egap_fit"
    ptype = float
    atype = "number"

Egap_fit = _Egap_fit()
    
class _Egap_type(Keyword):
    """band gap type (`mandatory`). Units: ``.
    
    

    Returns:
        str: Given a band gap, this keyword describes if the system is a metal, a semi-metal, an insulator with direct or indirect band gap.
        
    """
    name = "Egap_type"
    ptype = str
    atype = "string"

Egap_type = _Egap_type()
    
class _PV_atom(Keyword):
    """atomic pressure*volume (`mandatory`). Units: `eV/atom`.
    
    

    Returns:
        float: Pressure multiplied by volume of the atom.
        
    """
    name = "PV_atom"
    ptype = float
    atype = "number"

PV_atom = _PV_atom()
    
class _PV_cell(Keyword):
    """unit cell pressure*volume (`mandatory`). Units: `eV`.
    
    

    Returns:
        float: Pressure multiplied by volume of the unit cell.
        
    """
    name = "PV_cell"
    ptype = float
    atype = "number"

PV_cell = _PV_cell()
    
class _Pearson_symbol_orig(Keyword):
    """original Pearson symbol (`mandatory`). Units: ``.
    
    

    Returns:
        str: Returns the Pearson symbol of the original-unrelaxed structure before the calculation.
        
    """
    name = "Pearson_symbol_orig"
    ptype = str
    atype = "string"

Pearson_symbol_orig = _Pearson_symbol_orig()
    
class _Pearson_symbol_relax(Keyword):
    """relaxed Pearson symbol (`mandatory`). Units: ``.
    
    

    Returns:
        str: Returns the Pearson symbol of the relaxed structure after the calculation.
        
    """
    name = "Pearson_symbol_relax"
    ptype = str
    atype = "string"

Pearson_symbol_relax = _Pearson_symbol_relax()
    
class _Pulay_stress(Keyword):
    """Pulay Stress (`mandatory`). Units: `kbar`.
    
    .. warning:: This keyword is still listed as development level. Use it
      knowing that it is subject to change or removal.

    Returns:
        float: Returns a metric of the basis set inconsistency for the calculation.
        
    """
    name = "Pulay_stress"
    ptype = float
    atype = "number"

Pulay_stress = _Pulay_stress()
    
class _Pullay_stress(Keyword):
    """Pulay Stress (`mandatory`). Units: `kbar`.
    
    .. warning:: This keyword is still listed as development level. Use it
      knowing that it is subject to change or removal.

    Returns:
        float: Returns a metric of the basis set inconsistency for the calculation.
        
    """
    name = "Pullay_stress"
    ptype = float
    atype = "number"

Pullay_stress = _Pullay_stress()
    
class _ael_bulk_modulus_reuss(Keyword):
    """AEL Reuss bulk modulus (`optional`). Units: `GPa`.
    
    

    Returns:
        float: Returns the bulk modulus as calculated using the Reuss method with AEL.
        
    """
    name = "ael_bulk_modulus_reuss"
    ptype = float
    atype = "number"

ael_bulk_modulus_reuss = _ael_bulk_modulus_reuss()
    
class _ael_bulk_modulus_voigt(Keyword):
    """AEL Voigt bulk modulus (`optional`). Units: `GPa`.
    
    

    Returns:
        float: Returns the bulk modulus as calculated using the Voigt method with AEL.
        
    """
    name = "ael_bulk_modulus_voigt"
    ptype = float
    atype = "number"

ael_bulk_modulus_voigt = _ael_bulk_modulus_voigt()
    
class _ael_bulk_modulus_vrh(Keyword):
    """AEL VRH bulk modulus (`optional`). Units: `GPa`.
    
    

    Returns:
        float: Returns the bulk modulus as calculated using the Voigt-Reuss-Hill average with AEL.
        
    """
    name = "ael_bulk_modulus_vrh"
    ptype = float
    atype = "number"

ael_bulk_modulus_vrh = _ael_bulk_modulus_vrh()
    
class _ael_elastic_anistropy(Keyword):
    """AEL elastic anistropy (`optional`). Units: ``.
    
    

    Returns:
        float: Returns the elastic anistropy as calculated with AEL.
        
    """
    name = "ael_elastic_anistropy"
    ptype = float
    atype = "number"

ael_elastic_anistropy = _ael_elastic_anistropy()
    
class _ael_poisson_ratio(Keyword):
    """AEL Poisson ratio (`optional`). Units: ``.
    
    

    Returns:
        float: Returns the istropic Poisson ratio as calculated with AEL.
        
    """
    name = "ael_poisson_ratio"
    ptype = float
    atype = "number"

ael_poisson_ratio = _ael_poisson_ratio()
    
class _ael_shear_modulus_reuss(Keyword):
    """AEL Reuss shear modulus (`optional`). Units: `GPa`.
    
    

    Returns:
        float: Returns the shear modulus as calculated using the Reuss method with AEL.
        
    """
    name = "ael_shear_modulus_reuss"
    ptype = float
    atype = "number"

ael_shear_modulus_reuss = _ael_shear_modulus_reuss()
    
class _ael_shear_modulus_voigt(Keyword):
    """AEL Voigt shear modulus (`optional`). Units: `GPa`.
    
    

    Returns:
        float: Returns the shear modulus as calculated using the Voigt method with AEL.
        
    """
    name = "ael_shear_modulus_voigt"
    ptype = float
    atype = "number"

ael_shear_modulus_voigt = _ael_shear_modulus_voigt()
    
class _ael_shear_modulus_vrh(Keyword):
    """AEL VRH shear modulus (`optional`). Units: `GPa`.
    
    

    Returns:
        float: Returns the shear modulus as calculated using the Voigt-Reuss-Hill average with AEL.
        
    """
    name = "ael_shear_modulus_vrh"
    ptype = float
    atype = "number"

ael_shear_modulus_vrh = _ael_shear_modulus_vrh()
    
class _aflow_version(Keyword):
    """aflow version (`optional`). Units: ``.
    
    

    Returns:
        str: Returns the version number of AFLOW used to perform the calculation.
        
    """
    name = "aflow_version"
    ptype = str
    atype = "string"

aflow_version = _aflow_version()
    
class _aflowlib_date(Keyword):
    """material generation date (`optional`). Units: ``.
    
    

    Returns:
        str: Returns the date of the AFLOW post-processor which generated the entry for the library.
        
    """
    name = "aflowlib_date"
    ptype = str
    atype = "string"

aflowlib_date = _aflowlib_date()
    
class _aflowlib_entries(Keyword):
    """aflowlib entries (`conditional`). Units: ``.
    
    

    Returns:
        list: For projects and set-layer entries, aflowlib_entries lists the available sub-entries which are associated with the $aurl of the subdirectories.  By parsing $aurl/?aflowlib_entries (containing $aurl/aflowlib_entries_number entries) the user finds further locations to interrogate.
        
    """
    name = "aflowlib_entries"
    ptype = list
    atype = "strings"

aflowlib_entries = _aflowlib_entries()
    
class _aflowlib_entries_number(Keyword):
    """aflowlib entry count (`conditional`). Units: ``.
    
    

    Returns:
        float: For projects and set-layer entries, aflowlib_entrieslists the available sub-entries which are associated with the $aurl of the subdirectories.  By parsing $aurl/?aflowlib_entries (containing $aurl/aflowlib_entries_number entries) the user finds further locations to interrogate.
        
    """
    name = "aflowlib_entries_number"
    ptype = float
    atype = "number"

aflowlib_entries_number = _aflowlib_entries_number()
    
class _aflowlib_version(Keyword):
    """aflowlib version (`optional`). Units: ``.
    
    

    Returns:
        str: Returns the version of the AFLOW post-processor which generated the entry for the library.
        
    """
    name = "aflowlib_version"
    ptype = str
    atype = "string"

aflowlib_version = _aflowlib_version()
    
class _agl_acoustic_debye(Keyword):
    """AGL acoustic Debye temperature (`optional`). Units: `K`.
    
    

    Returns:
        float: Returns the acoustic Debye temperature as calculated with AGL.
        
    """
    name = "agl_acoustic_debye"
    ptype = float
    atype = "number"

agl_acoustic_debye = _agl_acoustic_debye()
    
class _agl_bulk_modulus_isothermal_300K(Keyword):
    """AGL isothermal bulk modulus 300K (`optional`). Units: `GPa`.
    
    

    Returns:
        float: Returns the isothermal bulk modulus at 300K as calculated with AGL.
        
    """
    name = "agl_bulk_modulus_isothermal_300K"
    ptype = float
    atype = "number"

agl_bulk_modulus_isothermal_300K = _agl_bulk_modulus_isothermal_300K()
    
class _agl_bulk_modulus_static_300K(Keyword):
    """AGL static bulk modulus 300K (`optional`). Units: `GPa`.
    
    

    Returns:
        float: Returns the static bulk modulus at 300K as calculated with AGL.
        
    """
    name = "agl_bulk_modulus_static_300K"
    ptype = float
    atype = "number"

agl_bulk_modulus_static_300K = _agl_bulk_modulus_static_300K()
    
class _agl_debye(Keyword):
    """AGL Debye temperature (`optional`). Units: `K`.
    
    

    Returns:
        float: Returns the Debye temperature as calculated with AGL.
        
    """
    name = "agl_debye"
    ptype = float
    atype = "number"

agl_debye = _agl_debye()
    
class _agl_gruneisen(Keyword):
    """AGL Gruneisen parameter (`optional`). Units: ``.
    
    

    Returns:
        float: Returns the Gruneisen parameter as calculated with AGL.
        
    """
    name = "agl_gruneisen"
    ptype = float
    atype = "number"

agl_gruneisen = _agl_gruneisen()
    
class _agl_heat_capacity_Cp_300K(Keyword):
    """AGL heat capacity Cp (`optional`). Units: `kB/cell`.
    
    

    Returns:
        float: Returns the heat capacity at constant pressure as calculated with AGL at 300K.
        
    """
    name = "agl_heat_capacity_Cp_300K"
    ptype = float
    atype = "number"

agl_heat_capacity_Cp_300K = _agl_heat_capacity_Cp_300K()
    
class _agl_heat_capacity_Cv_300K(Keyword):
    """AGL heat capacity Cv (`optional`). Units: `kB/cell`.
    
    

    Returns:
        float: Returns the heat capacity at constant volume as calculated with AGL at 300K.
        
    """
    name = "agl_heat_capacity_Cv_300K"
    ptype = float
    atype = "number"

agl_heat_capacity_Cv_300K = _agl_heat_capacity_Cv_300K()
    
class _agl_thermal_conductivity_300K(Keyword):
    """AGL thermal conductivity (`optional`). Units: `W/m*K`.
    
    

    Returns:
        float: Returns the thermal conductivity as calculated with AGL at 300K.
        
    """
    name = "agl_thermal_conductivity_300K"
    ptype = float
    atype = "number"

agl_thermal_conductivity_300K = _agl_thermal_conductivity_300K()
    
class _agl_thermal_expansion_300K(Keyword):
    """AGL thermal expansion (`optional`). Units: `1/K`.
    
    

    Returns:
        float: Returns the thermal expansion as calculated with AGL at 300K.
        
    """
    name = "agl_thermal_expansion_300K"
    ptype = float
    atype = "number"

agl_thermal_expansion_300K = _agl_thermal_expansion_300K()
    
class _auid(Keyword):
    """AFLOWLIB Unique Identifier (`mandatory`). Units: ``.
    
    

    Returns:
        str: AFLOWLIB Unique Identifier for the entry, AUID, which can be used as a publishable object identifier.
        
    """
    name = "auid"
    ptype = str
    atype = "string"

auid = _auid()
    
class _aurl(Keyword):
    """AFLOWLIB Uniform Resource Locator (`mandatory`). Units: ``.
    
    

    Returns:
        str: AFLOWLIB Uniform Resource Locator returns the AURL of the entry.
        
    """
    name = "aurl"
    ptype = str
    atype = "string"

aurl = _aurl()
    
class _author(Keyword):
    """author (`optional`). Units: ``.
    
    .. warning:: This keyword is still listed as development level. Use it
      knowing that it is subject to change or removal.

    Returns:
        list: Returns the name (not necessarily an individual) and affiliation associated with authorship of the data.
        
    """
    name = "author"
    ptype = list
    atype = "strings"

author = _author()
    
class _bader_atomic_volumes(Keyword):
    """atomic volume per atom (`optional`). Units: `&Aring;<sup>3</sup>`.
    
    

    Returns:
        list: Returns the volume of each atom of the primitive cell as calculated by the Bader Atoms in Molecules Analysis. This volume encapsulates the electron density associated with each atom above a threshold of 0.0001 electrons.
        
    """
    name = "bader_atomic_volumes"
    ptype = list
    atype = "numbers"

bader_atomic_volumes = _bader_atomic_volumes()
    
class _bader_net_charges(Keyword):
    """partial charge per atom (`optional`). Units: `electrons`.
    
    

    Returns:
        list: Returns a comma delimited set of partial charges per atom of the primitive cell as calculated by the Bader Atoms in Molecules Analysis.
        
    """
    name = "bader_net_charges"
    ptype = list
    atype = "numbers"

bader_net_charges = _bader_net_charges()
    
class _calculation_cores(Keyword):
    """used CPU cores (`optional`). Units: ``.
    
    

    Returns:
        float: Number of processors/cores used for the calculation.
        
    """
    name = "calculation_cores"
    ptype = float
    atype = "number"

calculation_cores = _calculation_cores()
    
class _calculation_memory(Keyword):
    """used RAM (`optional`). Units: `Megabytes`.
    
    

    Returns:
        float: The maximum memory used for the calculation.
        
    """
    name = "calculation_memory"
    ptype = float
    atype = "number"

calculation_memory = _calculation_memory()
    
class _calculation_time(Keyword):
    """used time (`optional`). Units: `seconds`.
    
    

    Returns:
        float: Total time taken for the calculation.
        
    """
    name = "calculation_time"
    ptype = float
    atype = "number"

calculation_time = _calculation_time()
    
class _catalog(Keyword):
    """catalog (`optional`). Units: ``.
    
    

    Returns:
        str: Returns the context set for the calculation.
        
    """
    name = "catalog"
    ptype = str
    atype = "string"

catalog = _catalog()
    
class _code(Keyword):
    """ab initio code (`optional`). Units: ``.
    
    

    Returns:
        str: Returns the software name and version used to perform the simulation.
        
    """
    name = "code"
    ptype = str
    atype = "string"

code = _code()
    
class _composition(Keyword):
    """composition (`optional`). Units: ``.
    
    

    Returns:
        list: Returns a comma delimited composition description of the structure entry in the calculated cell.
        
    """
    name = "composition"
    ptype = list
    atype = "numbers"

composition = _composition()
    
class _compound(Keyword):
    """chemical formula (`mandatory`). Units: ``.
    
    

    Returns:
        str: Returns the composition description of the compound in the calculated cell.
        
    """
    name = "compound"
    ptype = str
    atype = "string"

compound = _compound()
    
class _corresponding(Keyword):
    """coresponding (`optional`). Units: ``.
    
    .. warning:: This keyword is still listed as development level. Use it
      knowing that it is subject to change or removal.

    Returns:
        list: Returns the name (not necessarily an individual) and affiliation associated with the data origin concerning correspondence about data.
        
    """
    name = "corresponding"
    ptype = list
    atype = "strings"

corresponding = _corresponding()
    
class _data_api(Keyword):
    """REST API version (`mandatory`). Units: ``.
    
    

    Returns:
        str: AFLOWLIB version of the entry, API.}
        
    """
    name = "data_api"
    ptype = str
    atype = "string"

data_api = _data_api()
    
class _data_language(Keyword):
    """data language (`optional`). Units: ``.
    
    

    Returns:
        list: Gives the language of the data in AFLOWLIB.
        
    """
    name = "data_language"
    ptype = list
    atype = "strings"

data_language = _data_language()
    
class _data_source(Keyword):
    """data source (`optional`). Units: ``.
    
    

    Returns:
        list: Gives the source of the data in AFLOWLIB.
        
    """
    name = "data_source"
    ptype = list
    atype = "strings"

data_source = _data_source()
    
class _delta_electronic_energy_convergence(Keyword):
    """Electronic Energy of Convergence Step (`optional`). Units: ``.
    
    .. warning:: This keyword is still listed as development level. Use it
      knowing that it is subject to change or removal.

    Returns:
        float: Returns the change in energy from the last step of the convergence iteration.
        
    """
    name = "delta_electronic_energy_convergence"
    ptype = float
    atype = "number"

delta_electronic_energy_convergence = _delta_electronic_energy_convergence()
    
class _delta_electronic_energy_threshold(Keyword):
    """Electronic Energy of Convergence Threshold (`optional`). Units: ``.
    
    .. warning:: This keyword is still listed as development level. Use it
      knowing that it is subject to change or removal.

    Returns:
        float: Returns the maximimum change in energy required for the convergence iteration.
        
    """
    name = "delta_electronic_energy_threshold"
    ptype = float
    atype = "number"

delta_electronic_energy_threshold = _delta_electronic_energy_threshold()
    
class _density(Keyword):
    """mass density (`optional`). Units: `grams/cm<sup>3</sup>`.
    
    

    Returns:
        float: Returns the mass density in grams/cm3.
        
    """
    name = "density"
    ptype = float
    atype = "number"

density = _density()
    
class _dft_type(Keyword):
    """DFT type (`optional`). Units: ``.
    
    

    Returns:
        list: Returns information about the pseudopotential type, the exchange correlation functional used (normal or hybrid) and use of GW.
        
    """
    name = "dft_type"
    ptype = list
    atype = "strings"

dft_type = _dft_type()
    
class _eentropy_atom(Keyword):
    """atomistic electronic entropy (`optional`). Units: `eV/atom`.
    
    

    Returns:
        float: Returns the electronic entropy of the atom used to converge the ab initio calculation (smearing).
        
    """
    name = "eentropy_atom"
    ptype = float
    atype = "number"

eentropy_atom = _eentropy_atom()
    
class _eentropy_cell(Keyword):
    """unit cell electronic entropy (`optional`). Units: `eV/atom`.
    
    

    Returns:
        float: Returns the electronic entropy of the unit cell used to converge the ab initio calculation (smearing).
        
    """
    name = "eentropy_cell"
    ptype = float
    atype = "number"

eentropy_cell = _eentropy_cell()
    
class _energy_atom(Keyword):
    """atomic energy (`mandatory`). Units: `eV/atom`.
    
    

    Returns:
        float: Returns the total ab initio energy per atom- the value of energy_cell/$N$).
        
    """
    name = "energy_atom"
    ptype = float
    atype = "number"

energy_atom = _energy_atom()
    
class _energy_cell(Keyword):
    """unit cell energy (`mandatory`). Units: `eV`.
    
    

    Returns:
        float: Returns the total ab initio energy of the unit cell, E. At T=0K and p=0, this is the internal energy of the system (per unit cell).
        
    """
    name = "energy_cell"
    ptype = float
    atype = "number"

energy_cell = _energy_cell()
    
class _energy_cutoff(Keyword):
    """energy cutoff (`optional`). Units: `eV`.
    
    

    Returns:
        list: Set of energy cut-offs used during the various steps of the calculations.
        
    """
    name = "energy_cutoff"
    ptype = list
    atype = "numbers"

energy_cutoff = _energy_cutoff()
    
class _enthalpy_atom(Keyword):
    """atomic enthalpy (`mandatory`). Units: `eV/atom`.
    
    

    Returns:
        float: Returns the enthalpy per atom- the value of enthalpy_cell/N).
        
    """
    name = "enthalpy_atom"
    ptype = float
    atype = "number"

enthalpy_atom = _enthalpy_atom()
    
class _enthalpy_cell(Keyword):
    """unit cell enthalpy (`mandatory`). Units: `eV`.
    
    

    Returns:
        float: Returns the enthalpy of the system of the unit cell, H = E + PV.
        
    """
    name = "enthalpy_cell"
    ptype = float
    atype = "number"

enthalpy_cell = _enthalpy_cell()
    
class _enthalpy_formation_atom(Keyword):
    """atomic formation enthalpy (`mandatory`). Units: `eV/atom`.
    
    

    Returns:
        float: Returns the formation enthalpy DeltaHFatomic per atom).
        
    """
    name = "enthalpy_formation_atom"
    ptype = float
    atype = "number"

enthalpy_formation_atom = _enthalpy_formation_atom()
    
class _enthalpy_formation_cell(Keyword):
    """unit cell formation enthalpy (`mandatory`). Units: `eV`.
    
    

    Returns:
        float: Returns the formation enthalpy DeltaHF per unit cell.
        
    """
    name = "enthalpy_formation_cell"
    ptype = float
    atype = "number"

enthalpy_formation_cell = _enthalpy_formation_cell()
    
class _entropic_temperature(Keyword):
    """entropic temperature (`mandatory`). Units: `Kelvin`.
    
    

    Returns:
        float: Returns the entropic temperature for the structure.
        
    """
    name = "entropic_temperature"
    ptype = float
    atype = "number"

entropic_temperature = _entropic_temperature()
    
class _files(Keyword):
    """I/O files (`conditional`). Units: ``.
    
    

    Returns:
        list: Provides access to the input and output files used in the simulation (provenance data).
        
    """
    name = "files"
    ptype = list
    atype = "strings"

files = _files()
    
class _forces(Keyword):
    """Quantum Forces (`optional`). Units: `eV/&Aring;`.
    
    .. warning:: This keyword is still listed as development level. Use it
      knowing that it is subject to change or removal.

    Returns:
        numpy.ndarray: Final quantum mechanical forces (Fi,Fj,Fk) in the notation of the code.
        
    """
    name = "forces"
    ptype = numpy.ndarray
    atype = "numbers"

forces = _forces()
    
class _geometry(Keyword):
    """unit cell basis (`mandatory`). Units: `&Aring;`.
    
    

    Returns:
        list: Returns geometrical data describing the unit cell in the usual a,b,c,alpha,beta,gamma notation.
        
    """
    name = "geometry"
    ptype = list
    atype = "numbers"

geometry = _geometry()
    
class _keywords(Keyword):
    """Title (`mandatory`). Units: ``.
    
    .. warning:: This keyword is still listed as development level. Use it
      knowing that it is subject to change or removal.

    Returns:
        list: This includes the list of keywords available in the entry, separated by commas.
        
    """
    name = "keywords"
    ptype = list
    atype = "strings"

keywords = _keywords()
    
class _kpoints(Keyword):
    """K-point mesh (`optional`). Units: ``.
    
    

    Returns:
        dict: with keys ['relaxation', 'static', 'points', 'nsamples']
            describing the cells for the relaxation and static calculations, the
            k-space symmetry points of the structure and the number of samples.
    """
    name = "kpoints"
    ptype = dict
    atype = "numbers"

kpoints = _kpoints()
    
class _lattice_system_orig(Keyword):
    """original lattice system (`mandatory`). Units: ``.
    
    

    Returns:
        str: Return the lattice system and lattice variation (Brillouin zone) of the original-unrelaxed structure before the calculation.
        
    """
    name = "lattice_system_orig"
    ptype = str
    atype = "string"

lattice_system_orig = _lattice_system_orig()
    
class _lattice_system_relax(Keyword):
    """relaxed lattice system (`mandatory`). Units: ``.
    
    

    Returns:
        str: Return the lattice system and lattice variation (Brillouin zone) of the relaxed structure after the calculation.
        
    """
    name = "lattice_system_relax"
    ptype = str
    atype = "string"

lattice_system_relax = _lattice_system_relax()
    
class _lattice_variation_orig(Keyword):
    """original lattice variation (`mandatory`). Units: ``.
    
    

    Returns:
        str: Return the lattice system and lattice variation (Brillouin zone) of the original-unrelaxed structure before the calculation.
        
    """
    name = "lattice_variation_orig"
    ptype = str
    atype = "string"

lattice_variation_orig = _lattice_variation_orig()
    
class _lattice_variation_relax(Keyword):
    """relaxed lattice variation (`mandatory`). Units: ``.
    
    

    Returns:
        str: Return the lattice system and lattice variation (Brillouin zone) of the relaxed structure after the calculation.
        
    """
    name = "lattice_variation_relax"
    ptype = str
    atype = "string"

lattice_variation_relax = _lattice_variation_relax()
    
class _ldau_TLUJ(Keyword):
    """on site coulomb interaction (`mandatory`). Units: ``.
    
    .. warning:: This keyword is still listed as development level. Use it
      knowing that it is subject to change or removal.

    Returns:
        list: This vector of numbers contains the parameters of the DFT+U calculations, based on a corrective functional inspired by the Hubbard model.
        
    """
    name = "ldau_TLUJ"
    ptype = list
    atype = "numbers"

ldau_TLUJ = _ldau_TLUJ()
    
class _loop(Keyword):
    """process category (`optional`). Units: ``.
    
    

    Returns:
        list: Informs the user of the type of post-processing that was performed.
        
    """
    name = "loop"
    ptype = list
    atype = "strings"

loop = _loop()
    
class _natoms(Keyword):
    """number of atoms in unit cell (`mandatory`). Units: ``.
    
    

    Returns:
        float: Returns the number of atoms in the unit cell of the structure entry. The number can be non integer if partial occupation is considered within appropriate approximations.
        
    """
    name = "natoms"
    ptype = float
    atype = "number"

natoms = _natoms()
    
class _nbondxx(Keyword):
    """nearest neighbor bond lengths (`optional`). Units: `&Aring;`.
    
    

    Returns:
        list: Nearest neighbors bond lengths of the relaxed structure per ordered set of species Ai,Aj greater than or equal to i.
        
    """
    name = "nbondxx"
    ptype = list
    atype = "numbers"

nbondxx = _nbondxx()
    
class _node_CPU_Cores(Keyword):
    """available CPU cores (`optional`). Units: ``.
    
    

    Returns:
        float: Information about the number of cores in the node/cluster where the calculation was performed.
        
    """
    name = "node_CPU_Cores"
    ptype = float
    atype = "number"

node_CPU_Cores = _node_CPU_Cores()
    
class _node_CPU_MHz(Keyword):
    """CPU rate (`optional`). Units: `Megahertz`.
    
    

    Returns:
        float: Information about the CPU speed in the node/cluster where the calculation was performed.
        
    """
    name = "node_CPU_MHz"
    ptype = float
    atype = "number"

node_CPU_MHz = _node_CPU_MHz()
    
class _node_CPU_Model(Keyword):
    """CPU model (`optional`). Units: ``.
    
    

    Returns:
        str: Information about the CPU model in the node/cluster where the calculation was performed.
        
    """
    name = "node_CPU_Model"
    ptype = str
    atype = "string"

node_CPU_Model = _node_CPU_Model()
    
class _node_RAM_GB(Keyword):
    """available RAM (`optional`). Units: `Gigabytes`.
    
    

    Returns:
        float: Information about the RAM in the node/cluster where the calculation was performed.
        
    """
    name = "node_RAM_GB"
    ptype = float
    atype = "number"

node_RAM_GB = _node_RAM_GB()
    
class _nspecies(Keyword):
    """species count (`mandatory`). Units: ``.
    
    

    Returns:
        float: Returns the number of species in the system (e.g., binary = 2, ternary = 3, etc.).
        
    """
    name = "nspecies"
    ptype = float
    atype = "number"

nspecies = _nspecies()
    
class _positions_cartesian(Keyword):
    """relaxed absolute positions (`mandatory`). Units: `&Aring;`.
    
    .. warning:: This keyword is still listed as development level. Use it
      knowing that it is subject to change or removal.

    Returns:
        numpy.ndarray: Final Cartesian positions (xi,xj,xk) in the notation of the code.
        
    """
    name = "positions_cartesian"
    ptype = numpy.ndarray
    atype = "numbers"

positions_cartesian = _positions_cartesian()
    
class _positions_fractional(Keyword):
    """relaxed relative positions (`mandatory`). Units: ``.
    
    .. warning:: This keyword is still listed as development level. Use it
      knowing that it is subject to change or removal.

    Returns:
        numpy.ndarray: Final fractional positions (xi,xj,xk) with respect to the unit cell as specified in $geometry.
        
    """
    name = "positions_fractional"
    ptype = numpy.ndarray
    atype = "numbers"

positions_fractional = _positions_fractional()
    
class _pressure(Keyword):
    """external pressure (`mandatory`). Units: `kbar`.
    
    

    Returns:
        float: Returns the target pressure selected for the simulation.
        
    """
    name = "pressure"
    ptype = float
    atype = "number"

pressure = _pressure()
    
class _pressure_final(Keyword):
    """resulting pressure (`mandatory`). Units: `kbar`.
    
    .. warning:: This keyword is still listed as development level. Use it
      knowing that it is subject to change or removal.

    Returns:
        float: Returns the external pressure achieved by the simulation.
        
    """
    name = "pressure_final"
    ptype = float
    atype = "number"

pressure_final = _pressure_final()
    
class _pressure_residual(Keyword):
    """residual pressure (`mandatory`). Units: `kbar`.
    
    .. warning:: This keyword is still listed as development level. Use it
      knowing that it is subject to change or removal.

    Returns:
        float: Returns the external pressure achieved by the simulation.
        
    """
    name = "pressure_residual"
    ptype = float
    atype = "number"

pressure_residual = _pressure_residual()
    
class _prototype(Keyword):
    """original prototype (`mandatory`). Units: ``.
    
    

    Returns:
        str: Returns the AFLOW unrelaxed prototype which was used for the calculation.
        
    """
    name = "prototype"
    ptype = str
    atype = "string"

prototype = _prototype()
    
class _scintillation_attenuation_length(Keyword):
    """attenuation length (`mandatory`). Units: `cm`.
    
    

    Returns:
        float: Returns the scintillation attenuation length of the compound in cm.
        
    """
    name = "scintillation_attenuation_length"
    ptype = float
    atype = "number"

scintillation_attenuation_length = _scintillation_attenuation_length()
    
class _sg(Keyword):
    """space group of compound (`mandatory`). Units: ``.
    
    

    Returns:
        list: Evolution of the space group of the compound.  The first, second and third string represent space group name/number before the first, after the first, and after the last relaxation of the calculation.
        
    """
    name = "sg"
    ptype = list
    atype = "strings"

sg = _sg()
    
class _sg2(Keyword):
    """refined space group of compound  (`mandatory`). Units: ``.
    
    

    Returns:
        list: Evolution of the space group of the compound.  The first, second and third string represent space group name/number before the first, after the first, and after the last relaxation of the calculation.
        
    """
    name = "sg2"
    ptype = list
    atype = "strings"

sg2 = _sg2()
    
class _spacegroup_orig(Keyword):
    """original space group number (`mandatory`). Units: ``.
    
    

    Returns:
        float: Returns the spacegroup number of the original-unrelaxed structure before the calculation.
        
    """
    name = "spacegroup_orig"
    ptype = float
    atype = "number"

spacegroup_orig = _spacegroup_orig()
    
class _spacegroup_relax(Keyword):
    """relaxed space group number (`mandatory`). Units: ``.
    
    

    Returns:
        float: Returns the spacegroup number of the relaxed structure after the calculation.
        
    """
    name = "spacegroup_relax"
    ptype = float
    atype = "number"

spacegroup_relax = _spacegroup_relax()
    
class _species(Keyword):
    """atomic species (`mandatory`). Units: ``.
    
    

    Returns:
        list: Species of the atoms in this material.
        
    """
    name = "species"
    ptype = list
    atype = "strings"

species = _species()
    
class _species_pp(Keyword):
    """pseudopotential of chemical speciess (`mandatory`). Units: ``.
    
    

    Returns:
        list: Pseudopotentials of the atomic species.
        
    """
    name = "species_pp"
    ptype = list
    atype = "strings"

species_pp = _species_pp()
    
class _species_pp_ZVAL(Keyword):
    """valence atoms per species (`optional`). Units: `electrons`.
    
    

    Returns:
        list: Returns the number of valence electrons of the atomic species.
        
    """
    name = "species_pp_ZVAL"
    ptype = list
    atype = "numbers"

species_pp_ZVAL = _species_pp_ZVAL()
    
class _species_pp_version(Keyword):
    """pseudopotential version and species (`mandatory`). Units: ``.
    
    

    Returns:
        list: Species of the atoms, pseudopotentials species, and pseudopotential versions.
        
    """
    name = "species_pp_version"
    ptype = list
    atype = "strings"

species_pp_version = _species_pp_version()
    
class _spinD(Keyword):
    """spin decomposition over unit cell (`mandatory`). Units: `&mu;<sub>B</sub>`.
    
    

    Returns:
        list: For spin polarized calculations, the spin decomposition over the atoms of the cell.
        
    """
    name = "spinD"
    ptype = list
    atype = "numbers"

spinD = _spinD()
    
class _spinF(Keyword):
    """magnetization of unit cell at Fermi level (`mandatory`). Units: `&mu;<sub>B</sub>`.
    
    

    Returns:
        float: For spin polarized calculations, the magnetization of the cell at the Fermi level.
        
    """
    name = "spinF"
    ptype = float
    atype = "number"

spinF = _spinF()
    
class _spin_atom(Keyword):
    """atomic spin polarization (`mandatory`). Units: `&mu;<sub>B</sub>/atom`.
    
    

    Returns:
        float: For spin polarized calculations, the magnetization per atom.
        
    """
    name = "spin_atom"
    ptype = float
    atype = "number"

spin_atom = _spin_atom()
    
class _spin_cell(Keyword):
    """unit cell spin polarization (`mandatory`). Units: `&mu;<sub>B</sub>`.
    
    

    Returns:
        float: For spin polarized calculations, the total magnetization of the cell.
        
    """
    name = "spin_cell"
    ptype = float
    atype = "number"

spin_cell = _spin_cell()
    
class _sponsor(Keyword):
    """sponsor (`optional`). Units: ``.
    
    .. warning:: This keyword is still listed as development level. Use it
      knowing that it is subject to change or removal.

    Returns:
        list: Returns information about funding agencies and other sponsors for the data.
        
    """
    name = "sponsor"
    ptype = list
    atype = "strings"

sponsor = _sponsor()
    
class _stoich(Keyword):
    """unit cell stoichiometry (`optional`). Units: ``.
    
    .. warning:: This keyword is still listed as development level. Use it
      knowing that it is subject to change or removal.

    Returns:
        list: Similar to composition, returns a comma delimited stoichiometry description of the structure entry in the calculated cell.
        
    """
    name = "stoich"
    ptype = list
    atype = "numbers"

stoich = _stoich()
    
class _stoichiometry(Keyword):
    """unit cell stoichiometry (`mandatory`). Units: ``.
    
    

    Returns:
        list: Similar to composition, returns a comma delimited stoichiometry description of the structure entry in the calculated cell.
        
    """
    name = "stoichiometry"
    ptype = list
    atype = "numbers"

stoichiometry = _stoichiometry()
    
class _stress_tensor(Keyword):
    """Stress Tensor (`mandatory`). Units: ``.
    
    .. warning:: This keyword is still listed as development level. Use it
      knowing that it is subject to change or removal.

    Returns:
        list: Returns the stress tensor of the completed calculation.
        
    """
    name = "stress_tensor"
    ptype = list
    atype = "numbers"

stress_tensor = _stress_tensor()
    
class _valence_cell_iupac(Keyword):
    """unit cell IUPAC valence (`mandatory`). Units: ``.
    
    

    Returns:
        float: Returns IUPAC valence, the maximum number of univalent atoms that may combine with the atoms.
        
    """
    name = "valence_cell_iupac"
    ptype = float
    atype = "number"

valence_cell_iupac = _valence_cell_iupac()
    
class _valence_cell_std(Keyword):
    """unit cell standard valence (`mandatory`). Units: ``.
    
    

    Returns:
        float: Returns standard valence, the maximum number of univalent atoms that may combine with the atoms.
        
    """
    name = "valence_cell_std"
    ptype = float
    atype = "number"

valence_cell_std = _valence_cell_std()
    
class _volume_atom(Keyword):
    """atomic volume (`mandatory`). Units: `&Aring;<sup>3</sup>/atom`.
    
    

    Returns:
        float: Returns the volume per atom in the unit cell.
        
    """
    name = "volume_atom"
    ptype = float
    atype = "number"

volume_atom = _volume_atom()
    
class _volume_cell(Keyword):
    """unit cell volume (`mandatory`). Units: `&Aring;<sup>3</sup>`.
    
    

    Returns:
        float: Returns the volume of the unit cell.
        
    """
    name = "volume_cell"
    ptype = float
    atype = "number"

volume_cell = _volume_cell()




"""Functions for casting AFLOW type names to valid python objects. This also
creates :class:`numpy.ndarray` for vector or tensor-valued properties.
"""
import re
import numpy as np
from aflow import msg

_rx_int = re.compile(r"^\d+$")

def _strings(value):
    return list(value.split(','))

def _number(value):
    if _rx_int.match(value):
        return int(value)
    else:
        return float(value)

def _numbers(value):
    svals = list(value.split(','))
    dtype = type(_number(svals[0]))
    vals = list(map(dtype, svals))
    return np.array(vals)

def _forces(value):
    atoms = value.split(';')
    forces = [list(map(float, a.split(','))) for a in atoms]
    return np.array(forces)

def _kpoints(value):
    parts = value.split(';')
    relaxation = np.array(list(map(_number, parts[0].split(','))))
    static = np.array(list(map(_number, parts[1].split(','))))
    if len(parts) == 3: # pragma: no cover
        #The web page (possibly outdated) includes an example where
        #this would be the case. We include it here for
        #completeness. I haven't found a case yet that we could use in
        #the unit tests to trigger this.
        points = parts[-1].split('-')
        nsamples = None
    else:
        points = parts[-2].split('-')
        nsamples = int(parts[-1])
        
    return {
        "relaxation": relaxation,
        "static": static,
        "points": points,
        "nsamples": nsamples
        }

def _stoich(value):
    return list(map(_number, value.strip().split()))

docstrings = {
    "kpoints": """dict: with keys ['relaxation', 'static', 'points', 'nsamples']
describing the cells for the relaxation and static calculations, the
k-space symmetry points of the structure and the number of samples."""
}
"""dict: key-value pairs for custom docstrings describing the return
value of keywords with complex structure.
"""

exceptions = ["forces", "kpoints", "positions_cartesian",
              "positions_fractional", "spind", "stoich"]
"""list: of AFLOW keywords for which the casting has to be handled in a special
way.
"""

def ptype(atype, keyword):
    """Returns a `str` representing the *python* type for the
    specified AFLOW type and keyword.
    
    Args:
        atype (str): name of the AFLOW type.
        keyword (str): name of the keyword that the value is associated with.
    """
    castmap = {
        "string": "str",
        "strings": "list",
        "number": "float",
        "numbers": "list",
        "forces": "numpy.ndarray",
        "kpoints": "dict",
        "positions_cartesian": "numpy.ndarray",
        "positions_fractional": "numpy.ndarray",
        "spind": "list",
        "stoich": "list",
        "None": None,
        None: None
    }

    if keyword not in exceptions:
        return castmap[atype]
    else:
        return castmap[keyword]

def cast(atype, keyword, value):
    """Casts the specified value to a python type, using the AFLOW type as a
    reference.

    .. note:: Unfortunately, some of the AFLOW type names are not descriptive or
      unique enough to make general rule casting possible. Instead, we have to
      encode some exceptions directly in this module.
    
    Args:
        atype (str): name of the AFLOW type.
        keyword (str): name of the keyword that the value is associated with.
        value: object (usually a string) to cast into python types.
    """
    if value is None:
        return
    
    castmap = {
        "string": str,
        "strings": _strings,
        "number": _number,
        "numbers": _numbers,
        "forces": _forces,
        "kpoints": _kpoints,
        "positions_cartesian": _forces,
        "positions_fractional": _forces,
        "spind": _numbers,
        "stoich": _stoich,
        "None": lambda v: v,
        None: lambda v: v,
    }

    try:
        if keyword not in exceptions:
            return castmap[atype](value)
        else:
            return castmap[keyword](value)
    except:
        msg.err("Cannot cast {}; unknown format.".format(value))
    



"""Utility functions for interacting with file system, shell, etc.
"""
from os import path
from six import string_types
import six

def _get_reporoot():
    """Returns the absolute path to the repo root directory on the current
    system.
    """
    from os import path
    import aflow
    apath = path.abspath(aflow.__file__)
    return path.dirname(path.dirname(apath))

def load_module(modname, modpath, search_locs=None):
    """Loads the module specification and returns it as a python object.
    Args:
        modname (str): name that the module should be loaded under.
        modpath (str): full path to the code file in which the module is
          defined.
        search_locs (str): path to search for additional submodules.
    """
    if six.PY2:# pragma: no cover
        import imp
        result = imp.load_source(modname, modpath)
        return result
    else:
        from importlib.util import spec_from_file_location as specff
        from importlib.util import module_from_spec as modfs
        from importlib.machinery import SourceFileLoader
        spec = specff(modname, modpath, submodule_search_locations=search_locs)
        result = modfs(spec)
        spec.loader.exec_module(result)
        return result

reporoot = _get_reporoot()
"""The absolute path to the repo root on the local machine.
"""



"""Functions for controlling high-level search functionality for the
AFLOW database.
"""
server = "http://aflowlib.duke.edu/search/API/?"
"""str: API server address over HTTP.
"""

from aflow import msg
def search(catalog=None, batch_size=100):
    """Returns a :class:`aflow.control.Query` to help construct the search
    query.

    Args:
        catalog (str): one of the catalogs supported on AFLOW: ['icsd', 'lib1',
          'lib2', 'lib3']. Also supports a `list` of catalog names.
        batch_size (int): number of data entries to return per HTTP request.
    """
    return Query(catalog, batch_size)

class Query(object):
    """Represents a search againts the AFLUX API.

    Args:
        catalog (str): one of the catalogs supported on AFLOW: ['icsd', 'lib1',
          'lib2', 'lib3']. Also supports a `list` of catalog names.
        batch_size (int): number of data entries to return per HTTP request.
        step (int): step size over entries.

    Attributes:
        filters (list): of `str` filter arguments to pass to the matchbook
          section of the API request.
        select (list): of :class:`aflow.keywords.Keyword` to *include* in the request.
        excludes (list): of :class:`aflow.keywords.Keyword` to *exclude* in the request.
        orderby (str): name of the keyword to order by. AFLUX only supports a
          single order-by parameter for now.
        catalog (str): one of the catalogs supported on AFLOW: ['icsd', 'lib1',
          'lib2', 'lib3']. Also supports a `list` of catalog names.
        N (int): number of results in the current search query.
        reverse (bool): when True, reverse the order of the results in the
          query.
        k (int): number of datasets per page for the current iterator. Can be
          controlled by `batch_size`.
        responses (dict): keys are (n,k) tuples from the pagination; values are
          the corresponding JSON dictionaries.
        step (int): step size over entries.
    """
    def __init__(self, catalog=None, batch_size=100, step=1):
        from aflow.keywords import keywords
        self.filters = []
        self.selects = []#keywords]
        self.excludes = []
        self.order = None
        self.catalog = catalog if isinstance(catalog, (list, tuple, type(None))) else [catalog]
        self._N = None
        """int: number of results in the current search query."""
        self.reverse = False
        self._n = 1
        self.k = batch_size
        self.step = step
        self.responses = {}
        self._iter = 0
        """int: current integer id of the iterator in the *whole* dataset; this
        means it can have a value greater than :attr:`k`.
        """
        self._max_entry = None
        """int: index of the maximum entry that should be returned from this
        query.
        """
        self._matchbook = None
        """str: matchbook portion of the request URL.
        """
        self._final = False
        """bool: when True, this query is finalized and no additional filters,
        etc. can be added to it.
        """

    def reset_iter(self):
        """Resets the iterator back to zero so that the collection can be
        iterated over again *without* needing to request the data from the
        server again.
        """
        self._iter = 0
        
    @property
    def n(self):
        """Current page number for the iterator.
        """
        if self.reverse:
            return -1*self._n
        else:
            return self._n

    @property
    def N(self):
        if self._N is None:
            response = self._request(self.n, self.k)
        return self._N

    @property
    def max_N(self):
        """Returns the maximum integer index that will be reached by this query.
        """
        if self._max_entry is None:
            return self.N
        else:
            return self._max_entry
    
    def __len__(self):
        return self.N

    def __getitem__(self, seq):
        #We need to trigger the first request to make sure that the total
        #number of entries is fixed on the parent object, and so that the
        #pointers to the response caching are all the same.
        assert len(self) > 0
        
        from copy import copy
        result = copy(self)

        if type(seq) is slice:
            #Perform a shallow copy so that we get a new reference for the
            #indices.
            result._iter = seq.start if seq.start is not None else 0
            result._max_entry = seq.stop
            result.step = seq.step
            return result
        elif isinstance(seq, int):
            result._iter = seq
            return next(result)
        
    def _request(self, n, k):
        """Constructs the query string for this :class:`Query` object for the
        specified paging limits and then returns the response from the REST API
        as a python object.

        Args:
            n (int): page number of the results to return.
            k (int): number of datasets per page.
        """
        if len(self.responses) == 0:
            #We are making the very first request, finalize the query.
            self.finalize()
        
        import json
        from six.moves import urllib
        urlopen = urllib.request.urlopen
        url = "{0}{1},{2}".format(server, self.matchbook(),
                                  self._directives(n, k))
        rawresp = urlopen(url).read().decode("utf-8")
        try:
            response = json.loads(rawresp)
        except:# pragma: no cover
            #We can't easily simulate network failure...
            msg.err("{}\n\n{}".format(url, rawresp))
            return

        #If this is the first request, then save the number of results in the
        #query.
        if len(self.responses) == 0:
            self._N = int(next(iter(response.keys())).split()[-1])
        self.responses[n] = response

    def finalize(self):
        """Finalizes the current state of the query. This means that the request URL
        will be saved, but the individual keyword objects will be
        *reset*. Re-executing the search query will reconstruct the same object
        and request, but any cached responses will be lost.
        """
        #Generate the matchbook query, this has all the filters, selects and
        #ordering information.
        self.matchbook()
        #Next, reset all the keywords for the global AFLOW.
        from aflow.keywords import reset
        reset()
        #Switch out all of the keyword instances for their string
        #representations.
        self.selects = [str(s) for s in self.selects]
        self.filters = [str(f) for f in self.filters]
        self.excludes = [str(x) for x in self.excludes]
        self.order = str(self.order) if self.order is not None else None
        #Set the finalizer flag so that this object doesn't allow mutations.
        self._final = True
        
    def matchbook(self):
        """Constructs the matchbook portion of the query.
        """
        if not self._final or self._matchbook is None:
            items = []
            #AFLUX orders by the first element in the query. If we have an orderby
            #specified, then place it first.
            if self.order is not None:
                items.append(str(self.order))

            items.extend(list(map(str, self.selects)))
            items.extend(list(map(str, self.filters)))
            items.extend(["${}".format(str(k)) for k in self.excludes]) 
            self._matchbook = ','.join(items)

        return self._matchbook

    def _directives(self, n, k):
        """Returns the directives portion of the AFLUX query.

        Args:
            n (int): page number of the results to return.
            k (int): number of datasets per page.
        """
        items = []
        if self.catalog is not None:
            items.append("catalog({})".format(':'.join(self.catalog)))

        #Next, add the paging context. This query maintains its own paging
        #context
        items.append("paging({0:d},{1:d})".format(n, k))
        return ','.join(items)

    def __iter__(self):
        return self

    def next(self):# pragma: no cover
        """Yields a generator over AFLUX API request results.
        """
        return self.__next__()
    
    def __next__(self):
        """Yields a generator over AFLUX API request results.
        """
        #First, find out which entry we are on.
        n = (self._iter // self.k) + 1
        i = self._iter % self.k

        #Reverse the sign now that we have figured out the ordinal page number.
        if self.reverse:
            n *= -1

        if n not in self.responses and self._iter < self.max_N:
            self._n = abs(n)
            self._request(self.n, self.k)

        assert len(self.responses) > 0

        from aflow.entries import Entry
        if self._iter < self.max_N:
            index = self.k*(abs(n)-1) + i + 1
            key = "{} of {}".format(index, self.N)
            raw = self.responses[n][key]
            result = Entry(**raw)

            #Increment the iterator right before we return the entry.
            self._iter += 1
            return result
        else:
            raise StopIteration()

    def _final_check(self):
        """Checks whether this object is finalized; if it is, print a friendly
        message and return False, otherwise True.
        """
        if self._final:
            msg.info("This query has been finalized. It cannot be mutated. "
                     "Create a new query to change the matchbook.")
        return not self._final
        
    def filter(self, keyword):
        """Adds a search term to the current filter list. Calling :meth:`filter`
        multiple times will join the final filters together using logical *and*.

        Args:
            keyword (aflow.keywords.Keyword): that encapsulates the AFLUX
              request language logic.
        """
        if self._final_check():
            self._N = None
            self.filters.append(keyword)
        return self

    def select(self, *keywords):
        """Adds a keyword to the list of properties to return for each material
        in the request.

        Args:
            keywords (list): of :class:`aflow.keywords.Keyword` that
              encapsulates the AFLUX request language logic.
        """
        if self._final_check():
            self._N = None
            for keyword in keywords:
                if keyword is not self.order:
                    self.selects.append(keyword)
        return self

    def orderby(self, keyword, reverse=False):
        """Sets a keyword to be the one by which 

        Args:
            keyword (aflow.keywords.Keyword): that encapsulates the AFLUX
              request language logic.
            reverse (bool): when True, reverse the ordering.
        """
        if self._final_check():
            self._N = None
            self.order = keyword
            self.reverse = reverse
            if keyword in self.selects:
                self.selects.remove(keyword)
        return self
    
    def exclude(self, *keywords):
        """Sets a keyword to be *excluded* from the response.

        Args:
            keywords (list): of :class:`aflow.keywords.Keyword` that
              encapsulates the AFLUX request language logic.
        """
        if self._final_check():
            self._N = None
            for keyword in keywords:
                self.excludes.append(keyword)
        return self



"""Provides class and methods for abstracting the data from AFLOW into
python.
"""
from aflow.caster import cast
import aflow.keywords as kw

def _val_from_str(attr, value):
    """Retrieves the specified attribute's value, cast to an
    appropriate python type where possible.
    """
    clsname = "_{}".format(attr)
    if hasattr(kw, clsname):
        cls = getattr(kw, clsname)
        atype = getattr(cls, "atype")
        if attr == "kpoints":
            print(atype, attr, value)
        return cast(atype, attr, value)
    else:
        return value

class Entry(object):
    """Encapsulates the result of a single material entry in the AFLOW
    database.

    .. note:: Additional keyword values will be loaded lazily as
      requested (using additional HTTP requests). For optimization, it
      is recommended to request *all* known keywords up front.

    Args:
        kwargs (dict): of key-value pairs obtained from the initial
          AFLUX request.

    Attributes:
        attributes (dict): of key-value pairs requested for the given
          material. This will only be identical to the passed in the
          keyword arguments if no additional property requests have been
          made.
        raw (dict): original response dictionary (without any cast
          values).
    """
    def __init__(self, **kwargs):
        self.attributes = {a: _val_from_str(a, v) for a, v in kwargs.items()}
        self.raw = kwargs
        self._atoms = None
        """ase.atoms.Atoms: atoms object for the configuration in the
        database.
        """

    def __str__(self):
        aurl = self.attributes["aurl"].replace(".edu:", ".edu/")
        return "http://" + aurl
    def __eq__(self, other):
        return self.auid == other.auid
    def __hash__(self):
        return hash(self.auid)

    def _lazy_load(self, keyword):
        """Loads the value of the specified keyword via HTTP request against the
        AFLUX API, if it isn't already present on the object.

        Args:
            keyword (str): name of the keyword to retrieve for this entry.
        """
        if keyword in self.attributes:
            return self.attributes[keyword]
        else:
            import requests
            import json
            aurl = self.attributes["aurl"].replace(".edu:", ".edu/")
            url = "http://{0}?{1}".format(aurl, keyword)
            r = requests.get(url)

            if len(r.text) == 0:
                return

            #We need to coerce the string returned from aflow into the
            #appropriate python format.
            result = _val_from_str(keyword, r.text)
            self.attributes[keyword] = result
            return result

    def atoms(self, pattern="CONTCAR.relax*", quippy=False, keywords=None,
              calculator=None):
        """Creates a :class:`ase.atoms.Atoms` or a :class:`quippy.atoms.Atoms`
        object for this database entry.

        Args:
            pattern (str): pattern for choosing the file to generate the atomic
              lattice and positions from. The pattern is passed to
              :func:`~fnmatch.fnmatch` and the *last* entry in the list is
              returned (so that `CONTCAR.relax2` would be returned
              preferentially over `CONTCAR.relax1` or `CONTCAR.relax`).
            quippy (bool): when True, return a :class:`quippy.atoms.Atoms`
              object.
            keywords (dict): keys are keyword obects accessible from `aflow.K`;
              values are desired `str` names in the parameters dictionary of the
              atoms object.
            calculator (ase.calculators.Calculator): calculator to set for the
              newly created atoms object.

        Examples:
            Generate a :class:`quippy.atoms.Atoms` object and include the total
            energy and forces. Assume that `result` is a valid
            :func:`aflow.search` object.

            >>> entry = result[0] #Get the first result in the set.
            >>> keywords = {K.energy_cell: "dft_energy", K.forces: "dft_force"}
            >>> entry.atoms(quippy=True, keywords=keywords)
        """
        if self._atoms is not None:
            return self._atoms
        
        from fnmatch import fnmatch
        target = [f for f in self.files if fnmatch(f, pattern)][-1]
        aurl = self.attributes["aurl"].replace(".edu:", ".edu/")
        url = "http://{0}/{1}".format(aurl, target)

        import requests
        lines = requests.get(url).text.split('\n')
        preline = ' '.join(self.species).strip() + ' !'
        lines[0] = preline + lines[0]
        contcar = '\n'.join(lines)

        if quippy:# pragma: no cover
            import quippy
            reader = quippy.io.read
        else:
            from ase.io import read
            reader = read

        from six import StringIO
        cfile = StringIO(contcar)
        try:
            self._atoms = reader(cfile, format="vasp")
        finally:
            cfile.close()

        if calculator is not None:
            self._atoms.set_calculator(calculator)
        if keywords is None:
            return self._atoms

        self._atoms.results = {}
        for kw, pname in keywords.items():
            value = getattr(self, kw.name)
            if quippy: # pragma: no cover
                self._atoms.params.set_value(pname, value)
            else:
                #ASE only cares about certain values, but we'll save
                #them all anyway.
                self._atoms.results[pname] = value

        return self._atoms
        
    
    @property
    def Bravais_lattice_orig(self):
        """original bravais lattice (`optional`). Units: ``.
        
        
        

        Returns:
            str: Returns the Bravais lattice of the original unrelaxed structure before the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `Bravais_lattice_orig=MCLC`
        """
        return self._lazy_load("Bravais_lattice_orig")    
    
    @property
    def Bravais_lattice_relax(self):
        """relaxed bravais lattice (`optional`). Units: ``.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`forces`
          - :meth:`kpoints`
          - :meth:`stress_tensor`

        Returns:
            str: Returns the Bravais lattice of the original relaxed structure after the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `Bravais_lattice_relax=MCLC`
        """
        return self._lazy_load("Bravais_lattice_relax")    
    
    @property
    def Egap(self):
        """electronic energy band gap (`mandatory`). Units: `eV`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Band gap calculated with the approximations and pseudopotentials described by other keywords.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `Egap=2.5`
        """
        return self._lazy_load("Egap")    
    
    @property
    def Egap_fit(self):
        """fitted band gap (`mandatory`). Units: `eV`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Simple cross-validated correction (fit) of Egap.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `Egap_fit=3.5`
        """
        return self._lazy_load("Egap_fit")    
    
    @property
    def Egap_type(self):
        """band gap type (`mandatory`). Units: ``.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            str: Given a band gap, this keyword describes if the system is a metal, a semi-metal, an insulator with direct or indirect band gap.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `Egap_type=insulator_direct`
        """
        return self._lazy_load("Egap_type")    
    
    @property
    def PV_atom(self):
        """atomic pressure*volume (`mandatory`). Units: `eV/atom`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Pressure multiplied by volume of the atom.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `PV_atom=12.13`
        """
        return self._lazy_load("PV_atom")    
    
    @property
    def PV_cell(self):
        """unit cell pressure*volume (`mandatory`). Units: `eV`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Pressure multiplied by volume of the unit cell.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `PV_cell=12.13`
        """
        return self._lazy_load("PV_cell")    
    
    @property
    def Pearson_symbol_orig(self):
        """original Pearson symbol (`mandatory`). Units: ``.
        
        
        

        Returns:
            str: Returns the Pearson symbol of the original-unrelaxed structure before the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `Pearson_symbol_orig=mS32`
        """
        return self._lazy_load("Pearson_symbol_orig")    
    
    @property
    def Pearson_symbol_relax(self):
        """relaxed Pearson symbol (`mandatory`). Units: ``.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`stress_tensor`

        Returns:
            str: Returns the Pearson symbol of the relaxed structure after the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `Pearson_symbol_relax=mS32`
        """
        return self._lazy_load("Pearson_symbol_relax")    
    
    @property
    def Pulay_stress(self):
        """Pulay Stress (`mandatory`). Units: `kbar`.
        
        .. warning:: This keyword is still listed as development level. Use it
          knowing that it is subject to change or removal.
        

        Returns:
            float: Returns a metric of the basis set inconsistency for the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `pulay_stress=10.0`
        """
        return self._lazy_load("Pulay_stress")    
    
    @property
    def Pullay_stress(self):
        """Pulay Stress (`mandatory`). Units: `kbar`.
        
        .. warning:: This keyword is still listed as development level. Use it
          knowing that it is subject to change or removal.
        

        Returns:
            float: Returns a metric of the basis set inconsistency for the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `Pullay_stress=10.0`
        """
        return self._lazy_load("Pullay_stress")    
    
    @property
    def ael_bulk_modulus_reuss(self):
        """AEL Reuss bulk modulus (`optional`). Units: `GPa`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the bulk modulus as calculated using the Reuss method with AEL.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `ael_bulk_modulus_reuss=105.315`
        """
        return self._lazy_load("ael_bulk_modulus_reuss")    
    
    @property
    def ael_bulk_modulus_voigt(self):
        """AEL Voigt bulk modulus (`optional`). Units: `GPa`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the bulk modulus as calculated using the Voigt method with AEL.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `ael_bulk_modulus_voiht=105.315`
        """
        return self._lazy_load("ael_bulk_modulus_voigt")    
    
    @property
    def ael_bulk_modulus_vrh(self):
        """AEL VRH bulk modulus (`optional`). Units: `GPa`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the bulk modulus as calculated using the Voigt-Reuss-Hill average with AEL.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `ael_bulk_modulus_vrh=105.315`
        """
        return self._lazy_load("ael_bulk_modulus_vrh")    
    
    @property
    def ael_elastic_anistropy(self):
        """AEL elastic anistropy (`optional`). Units: ``.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the elastic anistropy as calculated with AEL.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `ael_elastic_anistropy=0.0008165`
        """
        return self._lazy_load("ael_elastic_anistropy")    
    
    @property
    def ael_poisson_ratio(self):
        """AEL Poisson ratio (`optional`). Units: ``.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the istropic Poisson ratio as calculated with AEL.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `ael_poisson_ratio=0.216`
        """
        return self._lazy_load("ael_poisson_ratio")    
    
    @property
    def ael_shear_modulus_reuss(self):
        """AEL Reuss shear modulus (`optional`). Units: `GPa`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the shear modulus as calculated using the Reuss method with AEL.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `ael_shear_modulus_reuss=73.787`
        """
        return self._lazy_load("ael_shear_modulus_reuss")    
    
    @property
    def ael_shear_modulus_voigt(self):
        """AEL Voigt shear modulus (`optional`). Units: `GPa`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the shear modulus as calculated using the Voigt method with AEL.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `ael_shear_modulus_voigt=73.799`
        """
        return self._lazy_load("ael_shear_modulus_voigt")    
    
    @property
    def ael_shear_modulus_vrh(self):
        """AEL VRH shear modulus (`optional`). Units: `GPa`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the shear modulus as calculated using the Voigt-Reuss-Hill average with AEL.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `ael_shear_modulus_vrh=73.793`
        """
        return self._lazy_load("ael_shear_modulus_vrh")    
    
    @property
    def aflow_version(self):
        """aflow version (`optional`). Units: ``.
        
        
        

        Returns:
            str: Returns the version number of AFLOW used to perform the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `aflow_version=aflow30641`
        """
        return self._lazy_load("aflow_version")    
    
    @property
    def aflowlib_date(self):
        """material generation date (`optional`). Units: ``.
        
        
        

        Returns:
            str: Returns the date of the AFLOW post-processor which generated the entry for the library.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `aflowlib_date=20140204_13:10:39_GMT-5`
        """
        return self._lazy_load("aflowlib_date")    
    
    @property
    def aflowlib_entries(self):
        """aflowlib entries (`conditional`). Units: ``.
        
        
        

        Returns:
            list: For projects and set-layer entries, aflowlib_entries lists the available sub-entries which are associated with the $aurl of the subdirectories.  By parsing $aurl/?aflowlib_entries (containing $aurl/aflowlib_entries_number entries) the user finds further locations to interrogate.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `aflowlib_entries=AgAl,AgAs,AgAu,AgB_h,AgBa_sv,AgBe_sv,AgBi_d,AgBr,AgCa_sv,...`
        """
        return self._lazy_load("aflowlib_entries")    
    
    @property
    def aflowlib_entries_number(self):
        """aflowlib entry count (`conditional`). Units: ``.
        
        
        

        Returns:
            float: For projects and set-layer entries, aflowlib_entrieslists the available sub-entries which are associated with the $aurl of the subdirectories.  By parsing $aurl/?aflowlib_entries (containing $aurl/aflowlib_entries_number entries) the user finds further locations to interrogate.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `aflowlib_entries_number=654`
        """
        return self._lazy_load("aflowlib_entries_number")    
    
    @property
    def aflowlib_version(self):
        """aflowlib version (`optional`). Units: ``.
        
        
        

        Returns:
            str: Returns the version of the AFLOW post-processor which generated the entry for the library.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `aflowlib_version=3.1.103`
        """
        return self._lazy_load("aflowlib_version")    
    
    @property
    def agl_acoustic_debye(self):
        """AGL acoustic Debye temperature (`optional`). Units: `K`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the acoustic Debye temperature as calculated with AGL.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `agl_acoustic_debye=492`
        """
        return self._lazy_load("agl_acoustic_debye")    
    
    @property
    def agl_bulk_modulus_isothermal_300K(self):
        """AGL isothermal bulk modulus 300K (`optional`). Units: `GPa`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the isothermal bulk modulus at 300K as calculated with AGL.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `agl_bulk_modulus_isothermal_300K=96.6`
        """
        return self._lazy_load("agl_bulk_modulus_isothermal_300K")    
    
    @property
    def agl_bulk_modulus_static_300K(self):
        """AGL static bulk modulus 300K (`optional`). Units: `GPa`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the static bulk modulus at 300K as calculated with AGL.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `agl_bulk_modulus_static_300K=99.6`
        """
        return self._lazy_load("agl_bulk_modulus_static_300K")    
    
    @property
    def agl_debye(self):
        """AGL Debye temperature (`optional`). Units: `K`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the Debye temperature as calculated with AGL.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `agl_debye=620`
        """
        return self._lazy_load("agl_debye")    
    
    @property
    def agl_gruneisen(self):
        """AGL Gruneisen parameter (`optional`). Units: ``.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the Gruneisen parameter as calculated with AGL.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `agl_gruneisen=2.06`
        """
        return self._lazy_load("agl_gruneisen")    
    
    @property
    def agl_heat_capacity_Cp_300K(self):
        """AGL heat capacity Cp (`optional`). Units: `kB/cell`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the heat capacity at constant pressure as calculated with AGL at 300K.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `agl_heat_capacity_Cp_300K=5.502`
        """
        return self._lazy_load("agl_heat_capacity_Cp_300K")    
    
    @property
    def agl_heat_capacity_Cv_300K(self):
        """AGL heat capacity Cv (`optional`). Units: `kB/cell`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the heat capacity at constant volume as calculated with AGL at 300K.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `agl_heat_capacity_Cv_300K=4.901`
        """
        return self._lazy_load("agl_heat_capacity_Cv_300K")    
    
    @property
    def agl_thermal_conductivity_300K(self):
        """AGL thermal conductivity (`optional`). Units: `W/m*K`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the thermal conductivity as calculated with AGL at 300K.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `agl_thermal_conductivity_300K=24.41`
        """
        return self._lazy_load("agl_thermal_conductivity_300K")    
    
    @property
    def agl_thermal_expansion_300K(self):
        """AGL thermal expansion (`optional`). Units: `1/K`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the thermal expansion as calculated with AGL at 300K.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `agl_thermal_expansion_300K=4.997e-05`
        """
        return self._lazy_load("agl_thermal_expansion_300K")    
    
    @property
    def auid(self):
        """AFLOWLIB Unique Identifier (`mandatory`). Units: ``.
        
        
        

        Returns:
            str: AFLOWLIB Unique Identifier for the entry, AUID, which can be used as a publishable object identifier.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `auid=aflow:e9c6d914c4b8d9ca`
        """
        return self._lazy_load("auid")    
    
    @property
    def aurl(self):
        """AFLOWLIB Uniform Resource Locator (`mandatory`). Units: ``.
        
        
        

        Returns:
            str: AFLOWLIB Uniform Resource Locator returns the AURL of the entry.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `aurl=aflowlib.duke.edu:AFLOWDATA/LIB3_RAW/Bi_dRh_pvTi_sv/T0003.ABC:LDAU2`
        """
        return self._lazy_load("aurl")    
    
    @property
    def author(self):
        """author (`optional`). Units: ``.
        
        .. warning:: This keyword is still listed as development level. Use it
          knowing that it is subject to change or removal.
        

        Returns:
            list: Returns the name (not necessarily an individual) and affiliation associated with authorship of the data.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `author=Marco_Buongiorno_Nardelli,Ohad_Levy,Jesus_Carrete`
        """
        return self._lazy_load("author")    
    
    @property
    def bader_atomic_volumes(self):
        """atomic volume per atom (`optional`). Units: `&Aring;<sup>3</sup>`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            list: Returns the volume of each atom of the primitive cell as calculated by the Bader Atoms in Molecules Analysis. This volume encapsulates the electron density associated with each atom above a threshold of 0.0001 electrons.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `bader_atomic_volumes=15.235,12.581,13.009`
        """
        return self._lazy_load("bader_atomic_volumes")    
    
    @property
    def bader_net_charges(self):
        """partial charge per atom (`optional`). Units: `electrons`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            list: Returns a comma delimited set of partial charges per atom of the primitive cell as calculated by the Bader Atoms in Molecules Analysis.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `bader_net_charges=0.125,0.125,-0.25`
        """
        return self._lazy_load("bader_net_charges")    
    
    @property
    def calculation_cores(self):
        """used CPU cores (`optional`). Units: ``.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Number of processors/cores used for the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `calculation_cores=32`
        """
        return self._lazy_load("calculation_cores")    
    
    @property
    def calculation_memory(self):
        """used RAM (`optional`). Units: `Megabytes`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: The maximum memory used for the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `calculation_memory=32`
        """
        return self._lazy_load("calculation_memory")    
    
    @property
    def calculation_time(self):
        """used time (`optional`). Units: `seconds`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Total time taken for the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `calculation_time=32`
        """
        return self._lazy_load("calculation_time")    
    
    @property
    def catalog(self):
        """catalog (`optional`). Units: ``.
        
        
        

        Returns:
            str: Returns the context set for the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `catalog=icsd`
        """
        return self._lazy_load("catalog")    
    
    @property
    def code(self):
        """ab initio code (`optional`). Units: ``.
        
        
        

        Returns:
            str: Returns the software name and version used to perform the simulation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `code=vasp.4.6.35`
        """
        return self._lazy_load("code")    
    
    @property
    def composition(self):
        """composition (`optional`). Units: ``.
        
        
        

        Returns:
            list: Returns a comma delimited composition description of the structure entry in the calculated cell.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `composition=2,6,6`
        """
        return self._lazy_load("composition")    
    
    @property
    def compound(self):
        """chemical formula (`mandatory`). Units: ``.
        
        
        

        Returns:
            str: Returns the composition description of the compound in the calculated cell.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `compound=Co2Er6Si6`
        """
        return self._lazy_load("compound")    
    
    @property
    def corresponding(self):
        """coresponding (`optional`). Units: ``.
        
        .. warning:: This keyword is still listed as development level. Use it
          knowing that it is subject to change or removal.
        

        Returns:
            list: Returns the name (not necessarily an individual) and affiliation associated with the data origin concerning correspondence about data.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `corresponding=M_Buongiorno_Nardelli_mbn@unt.edu`
        """
        return self._lazy_load("corresponding")    
    
    @property
    def data_api(self):
        """REST API version (`mandatory`). Units: ``.
        
        
        

        Returns:
            str: AFLOWLIB version of the entry, API.}
        
        Examples:
            You can expect the *content* of the result to be something like:

            `data_api=aapi1.0`
        """
        return self._lazy_load("data_api")    
    
    @property
    def data_language(self):
        """data language (`optional`). Units: ``.
        
        
        

        Returns:
            list: Gives the language of the data in AFLOWLIB.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `data_language=aflowlib`
        """
        return self._lazy_load("data_language")    
    
    @property
    def data_source(self):
        """data source (`optional`). Units: ``.
        
        
        

        Returns:
            list: Gives the source of the data in AFLOWLIB.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `data_source=aflowlib`
        """
        return self._lazy_load("data_source")    
    
    @property
    def delta_electronic_energy_convergence(self):
        """Electronic Energy of Convergence Step (`optional`). Units: ``.
        
        .. warning:: This keyword is still listed as development level. Use it
          knowing that it is subject to change or removal.
        

        Returns:
            float: Returns the change in energy from the last step of the convergence iteration.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `delta_electronic_energy_convergence=6.09588e-05`
        """
        return self._lazy_load("delta_electronic_energy_convergence")    
    
    @property
    def delta_electronic_energy_threshold(self):
        """Electronic Energy of Convergence Threshold (`optional`). Units: ``.
        
        .. warning:: This keyword is still listed as development level. Use it
          knowing that it is subject to change or removal.
        

        Returns:
            float: Returns the maximimum change in energy required for the convergence iteration.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `delta_electronic_energy_threshold=0.0001`
        """
        return self._lazy_load("delta_electronic_energy_threshold")    
    
    @property
    def density(self):
        """mass density (`optional`). Units: `grams/cm<sup>3</sup>`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`
          - :meth:`pressure_residual`
          - :meth:`stress_tensor`

        Returns:
            float: Returns the mass density in grams/cm3.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `density=7.76665`
        """
        return self._lazy_load("density")    
    
    @property
    def dft_type(self):
        """DFT type (`optional`). Units: ``.
        
        
        

        Returns:
            list: Returns information about the pseudopotential type, the exchange correlation functional used (normal or hybrid) and use of GW.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `dft_type=PAW_PBE,HSE06`
        """
        return self._lazy_load("dft_type")    
    
    @property
    def eentropy_atom(self):
        """atomistic electronic entropy (`optional`). Units: `eV/atom`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the electronic entropy of the atom used to converge the ab initio calculation (smearing).
        
        Examples:
            You can expect the *content* of the result to be something like:

            `eentropy_atom=0.0011`
        """
        return self._lazy_load("eentropy_atom")    
    
    @property
    def eentropy_cell(self):
        """unit cell electronic entropy (`optional`). Units: `eV/atom`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the electronic entropy of the unit cell used to converge the ab initio calculation (smearing).
        
        Examples:
            You can expect the *content* of the result to be something like:

            `eentropy_cell=0.0011`
        """
        return self._lazy_load("eentropy_cell")    
    
    @property
    def energy_atom(self):
        """atomic energy (`mandatory`). Units: `eV/atom`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`
          - :meth:`pressure_residual`
          - :meth:`stress_tensor`

        Returns:
            float: Returns the total ab initio energy per atom- the value of energy_cell/$N$).
        
        Examples:
            You can expect the *content* of the result to be something like:

            `energy_atom=-82.1656`
        """
        return self._lazy_load("energy_atom")    
    
    @property
    def energy_cell(self):
        """unit cell energy (`mandatory`). Units: `eV`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`
          - :meth:`pressure_residual`
          - :meth:`stress_tensor`

        Returns:
            float: Returns the total ab initio energy of the unit cell, E. At T=0K and p=0, this is the internal energy of the system (per unit cell).
        
        Examples:
            You can expect the *content* of the result to be something like:

            `energy_cell=-82.1656`
        """
        return self._lazy_load("energy_cell")    
    
    @property
    def energy_cutoff(self):
        """energy cutoff (`optional`). Units: `eV`.
        
        
        

        Returns:
            list: Set of energy cut-offs used during the various steps of the calculations.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `energy_cutoff=384.1,384.1,384.1`
        """
        return self._lazy_load("energy_cutoff")    
    
    @property
    def enthalpy_atom(self):
        """atomic enthalpy (`mandatory`). Units: `eV/atom`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`
          - :meth:`pressure_residual`
          - :meth:`stress_tensor`

        Returns:
            float: Returns the enthalpy per atom- the value of enthalpy_cell/N).
        
        Examples:
            You can expect the *content* of the result to be something like:

            `enthalpy_atom=-82.1656`
        """
        return self._lazy_load("enthalpy_atom")    
    
    @property
    def enthalpy_cell(self):
        """unit cell enthalpy (`mandatory`). Units: `eV`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`
          - :meth:`pressure_residual`
          - :meth:`stress_tensor`

        Returns:
            float: Returns the enthalpy of the system of the unit cell, H = E + PV.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `enthalpy_cell=-82.1656`
        """
        return self._lazy_load("enthalpy_cell")    
    
    @property
    def enthalpy_formation_atom(self):
        """atomic formation enthalpy (`mandatory`). Units: `eV/atom`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the formation enthalpy DeltaHFatomic per atom).
        
        Examples:
            You can expect the *content* of the result to be something like:

            `enthalpy_formation_atom=-33.1587`
        """
        return self._lazy_load("enthalpy_formation_atom")    
    
    @property
    def enthalpy_formation_cell(self):
        """unit cell formation enthalpy (`mandatory`). Units: `eV`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the formation enthalpy DeltaHF per unit cell.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `enthalpy_formation_cell=-33.1587`
        """
        return self._lazy_load("enthalpy_formation_cell")    
    
    @property
    def entropic_temperature(self):
        """entropic temperature (`mandatory`). Units: `Kelvin`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the entropic temperature for the structure.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `entropic_temperature=1072.1`
        """
        return self._lazy_load("entropic_temperature")    
    
    @property
    def files(self):
        """I/O files (`conditional`). Units: ``.
        
        
        

        Returns:
            list: Provides access to the input and output files used in the simulation (provenance data).
        
        Examples:
            You can expect the *content* of the result to be something like:

            `files=Bi_dRh_pv.33.cif,Bi_dRh_pv.33.png,CONTCAR.relax,CONTCAR.relax1,`
        """
        return self._lazy_load("files")    
    
    @property
    def forces(self):
        """Quantum Forces (`optional`). Units: `eV/&Aring;`.
        
        .. warning:: This keyword is still listed as development level. Use it
          knowing that it is subject to change or removal.
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            numpy.ndarray: Final quantum mechanical forces (Fi,Fj,Fk) in the notation of the code.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `forces=0,-0.023928,0.000197;0,0.023928,-0.000197;...`
        """
        return self._lazy_load("forces")    
    
    @property
    def geometry(self):
        """unit cell basis (`mandatory`). Units: `&Aring;`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`
          - :meth:`pressure_residual`
          - :meth:`stress_tensor`

        Returns:
            list: Returns geometrical data describing the unit cell in the usual a,b,c,alpha,beta,gamma notation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `geometry=18.82,18.82,18.82,32.41,32.41,32.41`
        """
        return self._lazy_load("geometry")    
    
    @property
    def keywords(self):
        """Title (`mandatory`). Units: ``.
        
        .. warning:: This keyword is still listed as development level. Use it
          knowing that it is subject to change or removal.
        

        Returns:
            list: This includes the list of keywords available in the entry, separated by commas.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `keywords=aurl,auid,loop,code,compound,prototype,nspecies,natoms,...`
        """
        return self._lazy_load("keywords")    
    
    @property
    def kpoints(self):
        """K-point mesh (`optional`). Units: ``.
        
        
        

        Returns:
            dict: Set of k-point meshes uniquely identifying the various steps of the calculations, e.g. relaxation, static and electronic band structure (specifying the k-space symmetry points of the structure).
        
        Examples:
            You can expect the *content* of the result to be something like:

            `kpoints=10,10,10;16,16,16;G-X-W-K-G-L-U-W-L-K+U-X`
        """
        return self._lazy_load("kpoints")    
    
    @property
    def lattice_system_orig(self):
        """original lattice system (`mandatory`). Units: ``.
        
        
        

        Returns:
            str: Return the lattice system and lattice variation (Brillouin zone) of the original-unrelaxed structure before the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `lattice_system_orig=rhombohedral`
        """
        return self._lazy_load("lattice_system_orig")    
    
    @property
    def lattice_system_relax(self):
        """relaxed lattice system (`mandatory`). Units: ``.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`forces`
          - :meth:`kpoints`
          - :meth:`stress_tensor`

        Returns:
            str: Return the lattice system and lattice variation (Brillouin zone) of the relaxed structure after the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `lattice_system_relax=rhombohedral`
        """
        return self._lazy_load("lattice_system_relax")    
    
    @property
    def lattice_variation_orig(self):
        """original lattice variation (`mandatory`). Units: ``.
        
        
        

        Returns:
            str: Return the lattice system and lattice variation (Brillouin zone) of the original-unrelaxed structure before the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `lattice_variation_orig=rhombohedral`
        """
        return self._lazy_load("lattice_variation_orig")    
    
    @property
    def lattice_variation_relax(self):
        """relaxed lattice variation (`mandatory`). Units: ``.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`forces`
          - :meth:`kpoints`
          - :meth:`stress_tensor`

        Returns:
            str: Return the lattice system and lattice variation (Brillouin zone) of the relaxed structure after the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `lattice_variation_relax=rhombohedral`
        """
        return self._lazy_load("lattice_variation_relax")    
    
    @property
    def ldau_TLUJ(self):
        """on site coulomb interaction (`mandatory`). Units: ``.
        
        .. warning:: This keyword is still listed as development level. Use it
          knowing that it is subject to change or removal.
        

        Returns:
            list: This vector of numbers contains the parameters of the DFT+U calculations, based on a corrective functional inspired by the Hubbard model.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `ldau_TLUJ=2;2,0,0;5,0,0;0,0,0`
        """
        return self._lazy_load("ldau_TLUJ")    
    
    @property
    def loop(self):
        """process category (`optional`). Units: ``.
        
        
        

        Returns:
            list: Informs the user of the type of post-processing that was performed.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `loop=thermodynamics,bands,magnetic`
        """
        return self._lazy_load("loop")    
    
    @property
    def natoms(self):
        """number of atoms in unit cell (`mandatory`). Units: ``.
        
        
        

        Returns:
            float: Returns the number of atoms in the unit cell of the structure entry. The number can be non integer if partial occupation is considered within appropriate approximations.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `natoms=12`
        """
        return self._lazy_load("natoms")    
    
    @property
    def nbondxx(self):
        """nearest neighbor bond lengths (`optional`). Units: `&Aring;`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`forces`
          - :meth:`kpoints`
          - :meth:`pressure_residual`
          - :meth:`stress_tensor`

        Returns:
            list: Nearest neighbors bond lengths of the relaxed structure per ordered set of species Ai,Aj greater than or equal to i.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `nbondxx=1.2599,1.0911,1.0911,1.7818,1.2599,1.7818`
        """
        return self._lazy_load("nbondxx")    
    
    @property
    def node_CPU_Cores(self):
        """available CPU cores (`optional`). Units: ``.
        
        
        

        Returns:
            float: Information about the number of cores in the node/cluster where the calculation was performed.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `node_CPU_Cores=12`
        """
        return self._lazy_load("node_CPU_Cores")    
    
    @property
    def node_CPU_MHz(self):
        """CPU rate (`optional`). Units: `Megahertz`.
        
        
        

        Returns:
            float: Information about the CPU speed in the node/cluster where the calculation was performed.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `node_CPU_MHz=12`
        """
        return self._lazy_load("node_CPU_MHz")    
    
    @property
    def node_CPU_Model(self):
        """CPU model (`optional`). Units: ``.
        
        
        

        Returns:
            str: Information about the CPU model in the node/cluster where the calculation was performed.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `node_CPU_Model=12`
        """
        return self._lazy_load("node_CPU_Model")    
    
    @property
    def node_RAM_GB(self):
        """available RAM (`optional`). Units: `Gigabytes`.
        
        
        

        Returns:
            float: Information about the RAM in the node/cluster where the calculation was performed.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `node_RAM_GB=12`
        """
        return self._lazy_load("node_RAM_GB")    
    
    @property
    def nspecies(self):
        """species count (`mandatory`). Units: ``.
        
        
        

        Returns:
            float: Returns the number of species in the system (e.g., binary = 2, ternary = 3, etc.).
        
        Examples:
            You can expect the *content* of the result to be something like:

            `nspecies=3`
        """
        return self._lazy_load("nspecies")    
    
    @property
    def positions_cartesian(self):
        """relaxed absolute positions (`mandatory`). Units: `&Aring;`.
        
        .. warning:: This keyword is still listed as development level. Use it
          knowing that it is subject to change or removal.
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`forces`
          - :meth:`kpoints`
          - :meth:`pressure_residual`
          - :meth:`stress_tensor`

        Returns:
            numpy.ndarray: Final Cartesian positions (xi,xj,xk) in the notation of the code.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `positions_cartesian=0,0,0;18.18438,0,2.85027;...`
        """
        return self._lazy_load("positions_cartesian")    
    
    @property
    def positions_fractional(self):
        """relaxed relative positions (`mandatory`). Units: ``.
        
        .. warning:: This keyword is still listed as development level. Use it
          knowing that it is subject to change or removal.
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`forces`
          - :meth:`kpoints`
          - :meth:`pressure_residual`
          - :meth:`stress_tensor`

        Returns:
            numpy.ndarray: Final fractional positions (xi,xj,xk) with respect to the unit cell as specified in $geometry.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `positions_fractional=0,0,0;0.25,0.25,0.25;...`
        """
        return self._lazy_load("positions_fractional")    
    
    @property
    def pressure(self):
        """external pressure (`mandatory`). Units: `kbar`.
        
        
        

        Returns:
            float: Returns the target pressure selected for the simulation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `pressure=10.0`
        """
        return self._lazy_load("pressure")    
    
    @property
    def pressure_final(self):
        """resulting pressure (`mandatory`). Units: `kbar`.
        
        .. warning:: This keyword is still listed as development level. Use it
          knowing that it is subject to change or removal.
        

        Returns:
            float: Returns the external pressure achieved by the simulation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `pressure_final=10.0`
        """
        return self._lazy_load("pressure_final")    
    
    @property
    def pressure_residual(self):
        """residual pressure (`mandatory`). Units: `kbar`.
        
        .. warning:: This keyword is still listed as development level. Use it
          knowing that it is subject to change or removal.
        

        Returns:
            float: Returns the external pressure achieved by the simulation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `pressure_residual=10.0`
        """
        return self._lazy_load("pressure_residual")    
    
    @property
    def prototype(self):
        """original prototype (`mandatory`). Units: ``.
        
        
        

        Returns:
            str: Returns the AFLOW unrelaxed prototype which was used for the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `prototype=T0001.A2BC`
        """
        return self._lazy_load("prototype")    
    
    @property
    def scintillation_attenuation_length(self):
        """attenuation length (`mandatory`). Units: `cm`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns the scintillation attenuation length of the compound in cm.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `scintillation_attenuation_length=2.21895`
        """
        return self._lazy_load("scintillation_attenuation_length")    
    
    @property
    def sg(self):
        """space group of compound (`mandatory`). Units: ``.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`forces`
          - :meth:`kpoints`
          - :meth:`stress_tensor`

        Returns:
            list: Evolution of the space group of the compound.  The first, second and third string represent space group name/number before the first, after the first, and after the last relaxation of the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `sg=Fm-3m#225,Fm-3m#225,Fm-3m#225`
        """
        return self._lazy_load("sg")    
    
    @property
    def sg2(self):
        """refined space group of compound  (`mandatory`). Units: ``.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`forces`
          - :meth:`kpoints`
          - :meth:`stress_tensor`

        Returns:
            list: Evolution of the space group of the compound.  The first, second and third string represent space group name/number before the first, after the first, and after the last relaxation of the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `sg2=Fm-3m#225,Fm-3m#225,Fm-3m#225`
        """
        return self._lazy_load("sg2")    
    
    @property
    def spacegroup_orig(self):
        """original space group number (`mandatory`). Units: ``.
        
        
        

        Returns:
            float: Returns the spacegroup number of the original-unrelaxed structure before the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `spacegroup_orig=225`
        """
        return self._lazy_load("spacegroup_orig")    
    
    @property
    def spacegroup_relax(self):
        """relaxed space group number (`mandatory`). Units: ``.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`forces`
          - :meth:`kpoints`
          - :meth:`stress_tensor`

        Returns:
            float: Returns the spacegroup number of the relaxed structure after the calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `spacegroup_relax=225`
        """
        return self._lazy_load("spacegroup_relax")    
    
    @property
    def species(self):
        """atomic species (`mandatory`). Units: ``.
        
        
        

        Returns:
            list: Species of the atoms in this material.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `species=Y,Zn,Zr`
        """
        return self._lazy_load("species")    
    
    @property
    def species_pp(self):
        """pseudopotential of chemical speciess (`mandatory`). Units: ``.
        
        
        

        Returns:
            list: Pseudopotentials of the atomic species.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `species_pp=Y,Zn,Zr`
        """
        return self._lazy_load("species_pp")    
    
    @property
    def species_pp_ZVAL(self):
        """valence atoms per species (`optional`). Units: `electrons`.
        
        
        

        Returns:
            list: Returns the number of valence electrons of the atomic species.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `species_pp_ZVAL=3`
        """
        return self._lazy_load("species_pp_ZVAL")    
    
    @property
    def species_pp_version(self):
        """pseudopotential version and species (`mandatory`). Units: ``.
        
        
        

        Returns:
            list: Species of the atoms, pseudopotentials species, and pseudopotential versions.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `species_pp_version=Y,Zn,Zr`
        """
        return self._lazy_load("species_pp_version")    
    
    @property
    def spinD(self):
        """spin decomposition over unit cell (`mandatory`). Units: `&mu;<sub>B</sub>`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            list: For spin polarized calculations, the spin decomposition over the atoms of the cell.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `spinD=0.236,0.236,-0.023,1.005`
        """
        return self._lazy_load("spinD")    
    
    @property
    def spinF(self):
        """magnetization of unit cell at Fermi level (`mandatory`). Units: `&mu;<sub>B</sub>`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: For spin polarized calculations, the magnetization of the cell at the Fermi level.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `spinF=0.410879`
        """
        return self._lazy_load("spinF")    
    
    @property
    def spin_atom(self):
        """atomic spin polarization (`mandatory`). Units: `&mu;<sub>B</sub>/atom`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: For spin polarized calculations, the magnetization per atom.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `spin_atom=2.16419`
        """
        return self._lazy_load("spin_atom")    
    
    @property
    def spin_cell(self):
        """unit cell spin polarization (`mandatory`). Units: `&mu;<sub>B</sub>`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: For spin polarized calculations, the total magnetization of the cell.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `spin_cell=2.16419`
        """
        return self._lazy_load("spin_cell")    
    
    @property
    def sponsor(self):
        """sponsor (`optional`). Units: ``.
        
        .. warning:: This keyword is still listed as development level. Use it
          knowing that it is subject to change or removal.
        

        Returns:
            list: Returns information about funding agencies and other sponsors for the data.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `sponsor=DOD_N000141310635,NIST_70NANB12H163`
        """
        return self._lazy_load("sponsor")    
    
    @property
    def stoich(self):
        """unit cell stoichiometry (`optional`). Units: ``.
        
        .. warning:: This keyword is still listed as development level. Use it
          knowing that it is subject to change or removal.
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            list: Similar to composition, returns a comma delimited stoichiometry description of the structure entry in the calculated cell.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `stoichiometry=0.5,0.25,0.25`
        """
        return self._lazy_load("stoich")    
    
    @property
    def stoichiometry(self):
        """unit cell stoichiometry (`mandatory`). Units: ``.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            list: Similar to composition, returns a comma delimited stoichiometry description of the structure entry in the calculated cell.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `stoichiometry=0.5,0.25,0.25`
        """
        return self._lazy_load("stoichiometry")    
    
    @property
    def stress_tensor(self):
        """Stress Tensor (`mandatory`). Units: ``.
        
        .. warning:: This keyword is still listed as development level. Use it
          knowing that it is subject to change or removal.
        

        Returns:
            list: Returns the stress tensor of the completed calculation.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `stress_tensor=-0.96,-0,-0,-0,-0.96,-0,-0,-0,-0.96`
        """
        return self._lazy_load("stress_tensor")    
    
    @property
    def valence_cell_iupac(self):
        """unit cell IUPAC valence (`mandatory`). Units: ``.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns IUPAC valence, the maximum number of univalent atoms that may combine with the atoms.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `valence_cell_iupac=22`
        """
        return self._lazy_load("valence_cell_iupac")    
    
    @property
    def valence_cell_std(self):
        """unit cell standard valence (`mandatory`). Units: ``.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`kpoints`

        Returns:
            float: Returns standard valence, the maximum number of univalent atoms that may combine with the atoms.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `valence_cell_std=22`
        """
        return self._lazy_load("valence_cell_std")    
    
    @property
    def volume_atom(self):
        """atomic volume (`mandatory`). Units: `&Aring;<sup>3</sup>/atom`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`forces`
          - :meth:`kpoints`
          - :meth:`pressure_residual`
          - :meth:`stress_tensor`

        Returns:
            float: Returns the volume per atom in the unit cell.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `volume_atom=100.984`
        """
        return self._lazy_load("volume_atom")    
    
    @property
    def volume_cell(self):
        """unit cell volume (`mandatory`). Units: `&Aring;<sup>3</sup>`.
        
        
        .. note:: The following verifications are available for this
          keyword. They are exposed as additional methods on this object.
          
          - :meth:`energy_cutoff`
          - :meth:`forces`
          - :meth:`kpoints`
          - :meth:`pressure_residual`
          - :meth:`stress_tensor`

        Returns:
            float: Returns the volume of the unit cell.
        
        Examples:
            You can expect the *content* of the result to be something like:

            `volume_cell=100.984`
        """
        return self._lazy_load("volume_cell")    
    
